/**
 * Copyright 2017 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


/**
 * Unit-tests for the monitoring interceptors
 *
 * Mocks the librdkafka APIs and triggers interceptor calls,
 * verifies interceptor-produced message content by protobufs.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#ifndef _MSC_VER
#include <dlfcn.h>
#endif

#define LIBRDKAFKA_STATICLIB
#include <librdkafka/rdkafka.h>

#include "../platform.h"
#include "monitoring.pb-c.h"
#include "../tools/print-json.h"


#define TEST_TOPIC "mytopic"
#define PARTITION_CNT 2


/**
 * @brief The current test state
 */
struct state {
        Monitoring__ClientType clienttype;
        rd_kafka_t *rk;
        rd_kafka_topic_t *rkt;
        rd_kafka_interceptor_f_on_acknowledgement_t *on_acknowledgement;
        void *on_acknowledgement_opaque;
        rd_kafka_interceptor_f_on_acknowledgement_t *on_consume;
        void *on_consume_opaque;
        int64_t next_sequence[PARTITION_CNT];
        int64_t latency;  /* Introduced latency */
        int do_shutdown;
        int saw_shutdown[PARTITION_CNT];
        char *session;
        int session_cnt;
        char *exp_clientid;
};

static struct state state;


/* Pretty colours */
#define CLR_0   "\033[0m"
#define CLR_RED "\033[31m"
#define CLR_GRN "\033[32m"
#define CLR_BLU "\033[34m"

#define TEST_FAIL(...) do {                                             \
                fprintf(stderr, CLR_RED "TEST FAILED at %s:%d %s:\n",   \
                        __FILE__, __LINE__, __FUNCTION__);              \
                fprintf(stderr, __VA_ARGS__);                           \
                fprintf(stderr, "\n");                                  \
                exit(1);                                                \
        } while (0)

#define TEST_ASSERT(expr,...) do {                                      \
                int _v = !!(expr);                                      \
                if (_v) break;                                          \
                fprintf(stderr, CLR_RED "ASSERT FAILED at %s:%d %s: " # expr ": ", \
                        __FILE__, __LINE__, __FUNCTION__);      \
                fprintf(stderr, __VA_ARGS__);                           \
                fprintf(stderr, "\n");                                  \
                exit(1);                                                \
        } while (0)

#define TEST_SAY(...) do {                                              \
                fprintf(stderr, CLR_BLU "%s:%d: " CLR_0,                \
                        __FUNCTION__, __LINE__);                        \
                fprintf(stderr, __VA_ARGS__);                           \
        } while (0)


#define MY_CLUSTER_ID  "fake_cluster_id"

static void handle_interceptor_message (const rd_kafka_message_t *rkm);


/**
 * @name Dynamic library loading
 *
 *
 */


/**
 * @brief Unload a plugin library
 */
static void unload_plugin (void *h) {
#ifdef _MSC_VER
        FreeLibrary((HMODULE)h);
#else
        dlclose(h);
#endif
}


/**
 * @brief Load a plugin library
 */
static void *load_plugin (const char *path) {
        void *h;
#ifdef _MSC_VER
        h = (void *)LoadLibraryA(path);
        if (!h)
                TEST_FAIL("LoadLibraryA(%s) failed", path);
#else
        h = dlopen(path, RTLD_NOW | RTLD_LOCAL);
        if (!h)
                TEST_FAIL("dlopen(%s) failed: %s", path, dlerror());
#endif
        return h;
}

/**
 * @brief Call the conf_init() method of a plugin library.
 */
static rd_kafka_resp_err_t call_conf_init (void *h, rd_kafka_conf_t *conf,
                                           void **plug_opaquep,
                                           char *errstr, size_t errstr_size) {
        rd_kafka_plugin_f_conf_init_t *func;

#ifdef _MSC_VER
        func = GetProcAddress((HMODULE)h, "conf_init");
#else
        func = dlsym(h, "conf_init");
#endif

        TEST_ASSERT(func, "conf_init() not found");

        return func(conf, plug_opaquep, errstr, errstr_size);
}


/**
 * @name librdkafka overloads
 *
 */
rd_kafka_resp_err_t
rd_kafka_interceptor_add_on_acknowledgement (rd_kafka_t *rk,
                                  const char *ic_name,
                                  rd_kafka_interceptor_f_on_acknowledgement_t *on_acknowledgement,
                                  void *ic_opaque) {
        TEST_SAY("set on_acknowledgement\n");
        if (state.on_acknowledgement)
                TEST_FAIL("on_acknowledgement() set twice");

        state.on_acknowledgement = on_acknowledgement;
        state.on_acknowledgement_opaque = ic_opaque;

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}


rd_kafka_resp_err_t
rd_kafka_interceptor_add_on_consume (rd_kafka_t *rk,
                                     const char *ic_name,
                                     rd_kafka_interceptor_f_on_consume_t *on_consume,
                                     void *ic_opaque) {

        TEST_SAY("set on_consume\n");
        if (state.on_consume)
                TEST_FAIL("on_consume() set twice");

        state.on_consume = on_consume;
        state.on_consume_opaque = ic_opaque;

        return RD_KAFKA_RESP_ERR_NO_ERROR;
}



char *rd_kafka_clusterid (rd_kafka_t *rk, int timeout_ms) {
        return ic_strdup(MY_CLUSTER_ID);
}

int64_t rd_kafka_message_timestamp (const rd_kafka_message_t *rkmessage,
                                    rd_kafka_timestamp_type_t *tstype) {
        *tstype = RD_KAFKA_TIMESTAMP_CREATE_TIME;
        return get_wallclock_timestamp() - state.latency;
}

int64_t rd_kafka_message_latency (const rd_kafka_message_t *rkmessage) {
        return state.latency * 1000;
}



int rd_kafka_produce (rd_kafka_topic_t *rkt, int32_t partition,
                      int msgflags,
                      void *payload, size_t len,
                      const void *key, size_t keylen,
                      void *msg_opaque) {
        rd_kafka_message_t rkm = {
                .rkt       = rkt,
                .partition = (partition == RD_KAFKA_PARTITION_UA ?
                              0 : partition),
                .payload   = payload,
                .len       = len,
                .key       = (void *)key,
                .key_len   = keylen,
                ._private  = msg_opaque
        };

        TEST_SAY("produce to topic %s\n", rd_kafka_topic_name(rkt));

        if (rkt != state.rkt) {
                /* This is the interceptor producing to C3 */
                handle_interceptor_message(&rkm);

        } else {
                /* Standard produce call from intercepted client,
                 * trigger on_acknowledgement() (even though this should
                 * be triggered from rd_kafka_poll()) */
                rd_kafka_resp_err_t err;

                err = state.on_acknowledgement(state.rk, &rkm,
                                               state.on_acknowledgement_opaque);

                if (err)
                        TEST_FAIL("on_acknowledgement() failed: %s",
                                  rd_kafka_err2str(err));
        }

        return 0;
}

/*
 * End of overloads
 */


/**
 * @brief Handle and verify an interceptor message
 */
static void handle_interceptor_message (const rd_kafka_message_t *rkm) {
        Monitoring__MonitoringMessage *m;
        int64_t now;

        TEST_SAY("intercepted interceptor message\n");

#define MAGIC_BYTE_PROTOBUF 127

        TEST_ASSERT(rkm->len > 1, "expected > 1 bytes, not %d", (int)rkm->len);
        TEST_ASSERT(((uint8_t *)rkm->payload)[0] == MAGIC_BYTE_PROTOBUF,
                    "expected magic byte 0x%x, not 0x%x",
                    (uint8_t)MAGIC_BYTE_PROTOBUF,
                    ((uint8_t *)rkm->payload)[0]);

        m = monitoring__monitoring_message__unpack(NULL,
                                                   rkm->len-1,
                                                   (uint8_t *)rkm->payload+1);
        if (!m)
                TEST_FAIL("Failed to unpack interceptor message of %zu bytes",
                          rkm->len);

        monitoring_message_print_json(m);

        /* Verify the message */
        TEST_ASSERT(m->clienttype == state.clienttype,
                    "expected clienttype %d, not %d",
                    state.clienttype, m->clienttype);
        TEST_ASSERT(m->type == MONITORING__MESSAGE_TYPE__NORMAL ||
                    m->type == MONITORING__MESSAGE_TYPE__HEARTBEAT,
                    "unexpected type %d", m->type);
        TEST_ASSERT(m->clusterid, "%p", m->clusterid);
        TEST_ASSERT(!strcmp(m->clusterid, MY_CLUSTER_ID), "%s", m->clusterid);

        TEST_ASSERT(m->partition >= 0 && m->partition < PARTITION_CNT,
                    "partition %"PRId32" out of range", m->partition);
        TEST_ASSERT(m->topic, "expected topic");
        TEST_ASSERT(!strcmp(m->topic, TEST_TOPIC),
                    "expected topic %s, not %s", TEST_TOPIC, m->topic);

        TEST_ASSERT(m->session, "expected session");

        if (m->sequence == 0) {
                if (!state.session || strcmp(state.session, m->session)) {
                        int i;

                        TEST_ASSERT(state.saw_shutdown[m->partition],
                                    "New session should only be seen in "
                                    "shutdown:ed state");

                        if (state.session)
                                free(state.session);
                        state.session_cnt++;
                        state.session = strdup(m->session);
                        TEST_SAY("New session %s\n", m->session);

                        /* Reset shutdown and next_sequence state
                         * for all partitions */
                        for (i = 0 ; i < PARTITION_CNT ; i++) {
                                state.saw_shutdown[i] = 0;
                                state.next_sequence[i] = 0;
                        }
                } else {
                        TEST_ASSERT(m->sequence ==
                                    state.next_sequence[m->partition],
                                    "Expected sequence %"PRId64", got %"PRId64,
                                    state.next_sequence[m->partition],
                                    m->sequence);
                }

        } else {
                TEST_ASSERT(m->sequence == state.next_sequence[m->partition],
                            "Expected sequence %"PRId64", got %"PRId64,
                            state.next_sequence[m->partition], m->sequence);

                TEST_ASSERT(m->session,
                            "session unset for sequence >0 (%"PRId64")",
                            m->sequence);
                TEST_ASSERT(state.session,
                            "state.session unset for sequence >0 (%"PRId64")",
                            m->sequence);
                TEST_ASSERT(!strcmp(state.session, m->session),
                            "session changed %s != %s",
                            state.session, m->session);
        }

        state.next_sequence[m->partition] = m->sequence+1;
        now = get_wallclock_timestamp();
        TEST_ASSERT(m->timestamp <= now &&
                    m->timestamp + (2*1000) > now,
                    "timestamp %"PRId64" is too far off current time %"PRId64,
                    m->timestamp, now);

        if (m->type == MONITORING__MESSAGE_TYPE__NORMAL) {
                TEST_ASSERT(m->count > 0, "expected count > 0");
                TEST_ASSERT(m->minlatency >= state.latency,
                            "expected minlatency >= %"PRId64", got %"PRId64,
                            state.latency, m->minlatency);
                TEST_ASSERT(m->maxlatency <= state.latency,
                            "expected maxnlatency >= %"PRId64", got %"PRId64,
                            state.latency, m->maxlatency);
                TEST_ASSERT(m->totallatency >= state.latency * m->count,
                            "totallatency should be a product of "
                            "state.latency %"PRId64" and message count %"PRId64,
                            state.latency, m->count);
        } else if (m->type == MONITORING__MESSAGE_TYPE__HEARTBEAT) {
                TEST_ASSERT(m->minlatency == 0, "!");
                TEST_ASSERT(m->maxlatency == 0, "!");
                TEST_ASSERT(m->count == 0, "!");
        }

        TEST_ASSERT(m->clientid, "expected clientid");
        TEST_ASSERT(!strcmp(m->clientid, state.exp_clientid),
                    "expected clientid %s, not %s",
                    state.exp_clientid, m->clientid);

        if (!m->shutdown)
                TEST_ASSERT(!state.saw_shutdown[m->partition],
                            "received non-shutdown message after "
                            "shutdown message");

        state.saw_shutdown[m->partition] = m->shutdown;

        monitoring__monitoring_message__free_unpacked(m, NULL);
}


static void state_init (void) {
        int i;

        memset(&state, 0, sizeof(state));

        /* Mark all partitions as shutdown */
        for (i = 0 ; i < PARTITION_CNT ; i++)
                state.saw_shutdown[i] = 1;
}


/**
 * @brief Close down clients and verify shutdown behaviour.
 */
static void state_close (void) {
        int i;

        rd_kafka_topic_destroy(state.rkt);
        state.do_shutdown = 1;
        rd_kafka_destroy(state.rk);

        for (i = 0 ; i < PARTITION_CNT ; i++) {
                TEST_ASSERT(state.next_sequence[i] > 0,
                            "expected at least one interception for "
                            "partition %d", i);
                TEST_ASSERT(state.saw_shutdown[i],
                            "no shutdown interceptor message "
                            "seen for partition %d", i);
        }

        if (state.session)
                free(state.session);
}


/**
 * @brief Unit-test of producer interceptor, overloading/mocking certain
 *        librdkafka functionality.
 */
static void ut_test_producer (void *h) {
        rd_kafka_conf_t *conf = rd_kafka_conf_new();
        char errstr[512];
        rd_kafka_resp_err_t err;

        TEST_SAY("================ Producer test ===============\n");

        state_init();
        state.latency = 20;
        state.clienttype = MONITORING__CLIENT_TYPE__PRODUCER;

        if (rd_kafka_conf_set(conf, "debug", "interceptor",
                              errstr, sizeof(errstr)) != RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(\"debug\") failed: %s", errstr);

        err = call_conf_init(h, conf, NULL, errstr, sizeof(errstr));
        if (err)
                TEST_FAIL("conf_init() failed: %s: %s\n",
                          rd_kafka_err2str(err), errstr);

        if (rd_kafka_conf_set(conf, "confluent.monitoring.interceptor.icdebug",
                              "true", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(\"icdebug\") failed: %s", errstr);

        if (rd_kafka_conf_set(conf,
                              "confluent.monitoring.interceptor.publishMs",
                              "100", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
            TEST_FAIL("conf_set(..publishMs) failed: %s", errstr);

        state.exp_clientid = "myclientid";
        if (rd_kafka_conf_set(conf, "client.id", "myclientid",
                              errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(client.id) failed: %s", errstr);


        TEST_SAY("Calling new()\n");
        state.rk = rd_kafka_new(RD_KAFKA_PRODUCER, conf,
                                errstr, sizeof(errstr));
        if (!state.rk)
                TEST_FAIL("rd_kafka_new() failed: %s", errstr);

        if (!state.on_acknowledgement)
                TEST_FAIL("on_acknowledgement() not registered");

        state.rkt = rd_kafka_topic_new(state.rk, TEST_TOPIC, NULL);

        /* Produce to two partitions */
        rd_kafka_produce(state.rkt, 0,
                         RD_KAFKA_MSG_F_COPY, "hipart0", 7, NULL, 0, NULL);

        rd_kafka_produce(state.rkt, 1,
                         RD_KAFKA_MSG_F_COPY, "hipart1", 7, NULL, 0, NULL);

        rd_kafka_poll(state.rk, 1000);

        state_close();
}



/**
 * @brief on_consume() caller mock.
 */
static void mock_consume (int partition) {
        rd_kafka_message_t rkm = {
                .rkt       = state.rkt,
                .partition = partition,
                .payload   = "hello",
                .len       = 5
        };
        rd_kafka_resp_err_t err;

        err = state.on_consume(state.rk, &rkm, state.on_consume_opaque);

        if (err)
                TEST_FAIL("on_consume() failed: %s",
                          rd_kafka_err2str(err));
}


/**
 * @brief Unit-test of consumer interceptor, overloading/mocking certain
 *        librdkafka functionality.
 */
static void ut_test_consumer (void *h) {
        rd_kafka_conf_t *conf = rd_kafka_conf_new();
        char errstr[512];
        rd_kafka_resp_err_t err;

        TEST_SAY("================ Consumer test ===============\n");

        state_init();
        state.latency = 50;
        state.clienttype = MONITORING__CLIENT_TYPE__CONSUMER;

        if (rd_kafka_conf_set(conf, "debug", "interceptor",
                              errstr, sizeof(errstr)) != RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(\"debug\") failed: %s", errstr);

        err = call_conf_init(h, conf, NULL, errstr, sizeof(errstr));
        if (err)
                TEST_FAIL("conf_init() failed: %s: %s\n",
                          rd_kafka_err2str(err), errstr);

        if (rd_kafka_conf_set(conf, "confluent.monitoring.interceptor.icdebug",
                              "true", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(\"icdebug\") failed: %s", errstr);

        if (rd_kafka_conf_set(conf,
                              "confluent.monitoring.interceptor.publishMs",
                              "100", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(..publishMs) failed: %s", errstr);

        if (rd_kafka_conf_set(conf,
                              "confluent.monitoring.interceptor.client.id",
                              "inner.client.id", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(..client.id) failed: %s", errstr);

        state.exp_clientid = "outer.client.id";
        if (rd_kafka_conf_set(conf,
                              "client.id",
                              state.exp_clientid, errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(..client.id) failed: %s", errstr);

        if (rd_kafka_conf_set(conf,
                              "confluent.monitoring.interceptor.sessiondurationms",
                              "100", errstr, sizeof(errstr)) !=
            RD_KAFKA_CONF_OK)
                TEST_FAIL("conf_set(..sessiondurationms) failed: %s", errstr);


        TEST_SAY("Calling new()\n");
        state.rk = rd_kafka_new(RD_KAFKA_CONSUMER, conf,
                                errstr, sizeof(errstr));
        if (!state.rk)
                TEST_FAIL("rd_kafka_new() failed: %s", errstr);

        if (!state.on_consume)
                TEST_FAIL("on_consume() not registered");

        state.rkt = rd_kafka_topic_new(state.rk, TEST_TOPIC, NULL);

        /* Fake consume messages from two partitions */
        mock_consume(0);
        mock_consume(1);
        mock_consume(1);
        mock_consume(0);

        rd_kafka_poll(state.rk, 1000);

        TEST_ASSERT(state.session_cnt <= (1000/100)*1.2 &&
                    state.session_cnt >= (1000/100)*0.8,
                    "expected a new session every 100ms which should "
                    "result in %d +-20%% sessions, but saw %d",
                    1000/100, state.session_cnt);

        state_close();
}


int main (int argc, char **argv) {
        void *h;

        h = load_plugin("../monitoring-interceptor.so");

        ut_test_producer(h);

        ut_test_consumer(h);

        unload_plugin(h);

        return 0;
}
